from typing import Optional, List

import discord
from cpl_core.configuration import ConfigurationABC
from cpl_core.database.context import DatabaseContextABC
from cpl_discord.command import DiscordCommandABC
from cpl_discord.service import DiscordBotServiceABC
from cpl_translation import TranslatePipe
from discord import app_commands
from discord.ext import commands
from discord.ext.commands import Context

from bot_core.abc.client_utils_service_abc import ClientUtilsServiceABC
from bot_core.abc.message_service_abc import MessageServiceABC
from bot_core.helper.command_checks import CommandChecks
from bot_core.logging.command_logger import CommandLogger
from bot_core.pipes.date_time_offset_pipe import DateTimeOffsetPipe
from bot_data.abc.server_repository_abc import ServerRepositoryABC
from bot_data.abc.user_joined_server_repository_abc import UserJoinedServerRepositoryABC
from bot_data.abc.user_joined_voice_channel_abc import UserJoinedVoiceChannelRepositoryABC
from bot_data.abc.user_repository_abc import UserRepositoryABC
from modules.permission.abc.permission_service_abc import PermissionServiceABC


class UserGroup(DiscordCommandABC):

    def __init__(
            self,
            config: ConfigurationABC,
            logger: CommandLogger,
            message_service: MessageServiceABC,
            bot: DiscordBotServiceABC,
            client_utils: ClientUtilsServiceABC,
            permissions: PermissionServiceABC,
            servers: ServerRepositoryABC,
            db: DatabaseContextABC,
            users: UserRepositoryABC,
            user_joined_servers: UserJoinedServerRepositoryABC,
            user_joined_voice_channel: UserJoinedVoiceChannelRepositoryABC,
            translate: TranslatePipe,
            date: DateTimeOffsetPipe
    ):
        DiscordCommandABC.__init__(self)

        self._config = config
        self._logger = logger
        self._message_service = message_service
        self._bot = bot
        self._client_utils = client_utils
        self._permissions = permissions
        self._servers = servers
        self._db = db
        self._users = users
        self._user_joined_servers = user_joined_servers
        self._user_joined_voice_channel = user_joined_voice_channel
        self._t = translate
        self._date = date

        self._logger.trace(__name__, f'Loaded command service: {type(self).__name__}')

    @commands.hybrid_group()
    @commands.guild_only()
    async def user(self, ctx: Context):
        pass

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    async def info(self, ctx: Context, member: Optional[discord.Member] = None, *, wait: int = None):
        self._logger.debug(__name__, f'Received command user-info {ctx}:{member},{wait}')

        is_mod = self._permissions.is_member_moderator(ctx.author)
        if member is not None and not is_mod:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            return

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.server_id)
        joins = self._user_joined_servers.get_user_joined_servers_by_user_id(user.user_id)

        embed = discord.Embed(
            title=member.name,
            description=member.name,
            color=int('ef9d0d', 16)
        )

        ontime = self._user_joined_voice_channel.get_user_joined_voice_channels_by_user_id(user.user_id).sum(
            lambda join: round((join.leaved_on - join.joined_on).total_seconds() / 3600, 2))

        embed.add_field(name=self._t.transform('modules.base.user.info.fields.id'), value=member.id)
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.name'), value=member.name)
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.discord_join'),
                        value=self._date.transform(member.created_at), inline=False)
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.last_join'),
                        value=self._date.transform(member.joined_at), inline=False)
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.xp'), value=str(user.xp))
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.ontime'), value=str(ontime))

        roles = ''
        for role in member.roles:
            roles += f'{role.name}\n'
        embed.add_field(name=self._t.transform('modules.base.user.info.fields.roles'), value=roles, inline=False)

        if is_mod or member == ctx.author:
            joins_string = ''
            for join in joins:
                joins_string += f'{self._date.transform(join.joined_on)}\n'
            embed.add_field(name=self._t.transform('modules.base.user.info.fields.joins'), value=joins_string)

        if is_mod or member == ctx.author:
            lefts_string = ''
            for join in joins:
                if join.leaved_on is None:
                    if lefts_string == '':
                        lefts_string = '/'
                    continue
                lefts_string += f'{self._date.transform(join.leaved_on)}\n'

            embed.add_field(name=self._t.transform('modules.base.user.info.fields.lefts'), value=lefts_string)

        if is_mod or member == ctx.author:
            embed.add_field(name=self._t.transform('modules.base.user.info.fields.warnings'),
                            value=self._t.transform('common.not_implemented_yet'), inline=False)

        # send to interaction because of sensitive data
        await self._message_service.send_interaction_msg(ctx.interaction, embed, wait_before_delete=wait)
        self._logger.trace(__name__, f'Finished user-info command')

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    async def get(self, ctx: Context, atr: str, member: discord.Member = None):
        self._logger.debug(__name__, f'Received command user-get {ctx}:{member}')

        is_mod = self._permissions.is_member_moderator(ctx.author)
        if member is not None and not is_mod:
            await self._message_service.send_ctx_msg(ctx, self._t.transform('common.no_permission_message'))
            return

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.server_id)

        match atr:
            case 'xp':
                value = str(user.xp)
            
            case 'ontime':
                value = str(round(
                    self._user_joined_voice_channel.get_user_joined_voice_channels_by_user_id(user.user_id)
                    .sum(lambda join: (join.leaved_on - join.joined_on).total_seconds() / 3600),
                    2
                ))

            case other:
                await self._message_service.send_interaction_msg(ctx.interaction, self._t.transform('modules.base.user.get.atr_not_found').format(atr))
                return

        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f'modules.base.user.get.{atr}').format(member.mention, value)
        )

    @get.autocomplete('atr')
    async def get_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        atr_list = ['xp', 'ontime']
        return [app_commands.Choice(name=atr, value=atr) for atr in atr_list]

    @user.command()
    @commands.guild_only()
    @CommandChecks.check_is_ready()
    @CommandChecks.check_is_member_moderator()
    async def set(self, ctx: Context, atr: str, value: str, member: discord.Member = None):
        self._logger.debug(__name__, f'Received command user-set {ctx}:{member}')

        if value == '':
            await self._message_service.send_interaction_msg(ctx.interaction, self._t.transform('modules.base.user.set.error.no_value'))
            return

        if member is None or not isinstance(member, discord.Member):
            member = ctx.author

        server = self._servers.find_server_by_discord_id(ctx.guild.id)
        user = self._users.find_user_by_discord_id_and_server_id(member.id, server.server_id)

        match atr:
            case 'xp':
                try:
                    user.xp = int(value)
                    self._users.update_user(user)
                    self._db.save_changes()
                except Exception as e:
                    await self._logger.trace(__name__, f'Value couldn\'t be converted to int\n'+e)
                    # ToDo: Add text for this exception
                    await self._message_service.send_interaction_msg(ctx.interaction, self._t.transform('modules.base.user.error.value_type'))
                    return

            case other:
                # ToDo: Move atr_not_found
                await self._message_service.send_interaction_msg(ctx.interaction, self._t.transform(
                    'modules.base.user.error.atr_not_found').format(atr))
                return

        # ToDo: Add text for set
        await self._message_service.send_interaction_msg(
            ctx.interaction,
            self._t.transform(f'modules.base.user.set.{atr}').format(member.mention, value)
        )

    @set.autocomplete('atr')
    async def set_autocomplete(self, interaction: discord.Interaction, current: str) -> List[app_commands.Choice[str]]:
        atr_list = ['xp']
        return [app_commands.Choice(name=atr, value=atr) for atr in atr_list]
